from plone.tiles.interfaces import ESI_HEADER
from plone.tiles.interfaces import ESI_HEADER_KEY
from plone.tiles.interfaces import IESIRendered
from plone.tiles.interfaces import ITileType
from plone.tiles.tile import PersistentTile
from plone.tiles.tile import Tile
from Products.Five import BrowserView
from zExceptions import Unauthorized
from zope.component import queryUtility
from zope.interface import implementer

import os
import re
import transaction


try:
    from AccessControl.security import checkPermission
except ImportError:
    from zope.security import checkPermission


X_FRAME_OPTIONS = os.environ.get("PLONE_X_FRAME_OPTIONS", "SAMEORIGIN")

HEAD_CHILDREN = re.compile(r"<head[^>]*>(.*)</head>", re.I | re.S)
BODY_CHILDREN = re.compile(r"<body[^>]*>(.*)</body>", re.I | re.S)

ESI_NAMESPACE_MAP = {"esi": "http://www.edge-delivery.org/esi/1.0"}
_ESI_HREF = 'href="{url}/@@{esiMode}?{queryString}"'
ESI_TEMPLATE = (
    """\
<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN"
    "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">
<html xmlns="http://www.w3.org/1999/xhtml">
    <body>
        <a class="_esi_placeholder" rel="esi" """
    + _ESI_HREF
    + """></a>
    </body>
</html>
"""
)


def substituteESILinks(rendered):
    """Turn ESI links like <a class="_esi_placeholder" rel="esi" href="..." />
    into <esi:include /> links.

    ``rendered`` should be an HTML string.
    """

    rendered = re.sub(
        r"<html", '<html xmlns:esi="{}"'.format(ESI_NAMESPACE_MAP["esi"]), rendered, 1
    )
    return re.sub(
        r'<a class="_esi_placeholder" rel="esi" href="([^"]+)"></a>',
        r'<esi:include src="\1" />',
        rendered,
    )


class ConditionalESIRendering:
    head = False

    def render(self):
        raise NotImplementedError(
            "Override render() or set a class variable 'index' to point to "
            "a view page template file"
        )

    def __call__(self, *args, **kwargs):
        if self.request.getHeader(ESI_HEADER, "false").lower() == "true":
            mode = "esi-body"
            if self.head:
                mode = "esi-head"
            return ESI_TEMPLATE.format(
                url=(
                    self.request.get("PATH_INFO")
                    and self.request.get("PATH_INFO").replace(" ", "%20")
                    or self.request.getURL()
                ),
                queryString=self.request.get("QUERY_STRING", ""),
                esiMode=mode,
            )
        # Do not hide AttributeError inside index()
        try:
            self.index
        except AttributeError:
            return self.render()
        return self.index(*args, **kwargs)


# Convenience base classes


@implementer(IESIRendered)
class ESITile(ConditionalESIRendering, Tile):
    """Convenience class for tiles using ESI rendering.

    Set ``head`` to True if this tile renders <head /> content. The
    default is to render <body /> content.
    """

    head = False


@implementer(IESIRendered)
class ESIPersistentTile(ConditionalESIRendering, PersistentTile):
    """Convenience class for tiles using ESI rendering.

    Set ``head`` to True if this tile renders <head /> content. The
    default is to render <body /> content.
    """

    head = False


# ESI views


class ESIHead(BrowserView):
    """Render the head portion of a tile independently."""

    def __call__(self):
        """Return the children of the <head> tag as a fragment."""
        # Check for the registered view permission
        try:
            type_ = queryUtility(ITileType, self.context.__name__)
            permission = type_.view_permission
        except AttributeError:
            permission = None
        if permission:
            if not checkPermission(permission, self.context):
                raise Unauthorized()

        if self.request.getHeader(ESI_HEADER):
            del self.request.environ[ESI_HEADER_KEY]

        document = self.context()  # render the tile

        # Disable the theme so we don't <html/>-wrapped
        self.request.response.setHeader("X-Theme-Disabled", "1")

        match = HEAD_CHILDREN.search(document)
        if not match:
            return document
        return match.group(1).strip()


class ESIBody(BrowserView):
    """Render the head portion of a tile independently."""

    def __call__(self):
        """Return the children of the <head> tag as a fragment."""
        # Check for the registered view permission
        try:
            type_ = queryUtility(ITileType, self.context.__name__)
            permission = type_.view_permission
        except AttributeError:
            permission = None
        if permission:
            if not checkPermission(permission, self.context):
                raise Unauthorized()

        if self.request.getHeader(ESI_HEADER):
            del self.request.environ[ESI_HEADER_KEY]

        document = self.context()  # render the tile

        # Disable the theme so we don't <html/>-wrapped
        self.request.response.setHeader("X-Theme-Disabled", "1")

        match = BODY_CHILDREN.search(document)
        if not match:
            return document
        return match.group(1).strip()


class ESIProtectTransform:
    """Replacement transform for plone.protect's ProtectTransform,
    because ESI tile responses' HTML should not be transformed to
    avoid wrapping them with <html>-tag
    """

    order = 9000

    def __init__(self, published, request):
        self.published = published
        self.request = request

    def transform(self, result, encoding):
        from plone.protect.interfaces import IDisableCSRFProtection

        # clickjacking protection from plone.protect
        if X_FRAME_OPTIONS:
            if not self.request.response.getHeader("X-Frame-Options"):
                self.request.response.setHeader("X-Frame-Options", X_FRAME_OPTIONS)
        # drop X-Tile-Url
        if "x-tile-url" in self.request.response.headers:
            del self.request.response.headers["x-tile-url"]
        # ESI requests are always GET request and should not mutate DB
        # unless they provide IDisableCSRFProtection
        if not IDisableCSRFProtection.providedBy(self.request):
            transaction.abort()
        return None

    def transformBytes(self, result, encoding):
        return self.transform(result, encoding)

    def transformUnicode(self, result, encoding):
        return self.transform(result, encoding)

    def transformIterable(self, result, encoding):
        return self.transform(result, encoding)



from plone.rfc822.interfaces import IPrimaryField
from plone.tiles.data import decode
from plone.tiles.data import encode
from plone.tiles.testing import PLONE_TILES_INTEGRATION_TESTING
from zope import schema
from zope.interface import alsoProvides
from zope.interface import Interface

import unittest


class IQuerySchema(Interface):

    query = schema.List(
        title="Search terms",
        value_type=schema.Dict(value_type=schema.Field(), key_type=schema.TextLine()),
        required=False,
    )

    lines = schema.List(title="Strings", value_type=schema.TextLine(), required=False)

    title = schema.TextLine(title="Title")


class IWords(Interface):

    words = schema.List(title="Words", value_type=schema.TextLine(), required=False)


class IPrimary(Interface):

    words = schema.List(title="Words", value_type=schema.TextLine(), required=False)


alsoProvides(IPrimary["words"], IPrimaryField)


class TestEncode(unittest.TestCase):

    layer = PLONE_TILES_INTEGRATION_TESTING

    def test_encode_querystring_special(self):
        data = {
            "query": [
                {
                    "i": "Subject",
                    "o": "plone.app.querystring.operation.selection.any",
                    "v": ["äüö"],
                }
            ],
            "title": "Hello World",
        }
        self.assertEqual(
            encode(data, schema=IQuerySchema),
            (
                "query.i%3Arecords=Subject&query.o%3A"
                "records=plone.app.querystring.operation.selection.any&"
                "query.v%3Alist%3Arecords=%C3%A4%C3%BC%C3%B6&title=Hello+World"
            ),
        )

    def test_encode_unicode_lines(self):
        data = {"words": ["ä", "ö"]}
        self.assertEqual(
            encode(data, schema=IWords), "words%3Alist=%C3%A4&words%3Alist=%C3%B6"
        )

    def test_skip_encoding_primary_fields(self):
        data = {"words": ["ä", "ö"]}
        self.assertEqual(encode(data, schema=IPrimary), "")


class TestDecode(unittest.TestCase):

    layer = PLONE_TILES_INTEGRATION_TESTING

    def test_decode_unicode_lines(self):
        data = {"words": ["ä", "ö"]}
        self.assertEqual(decode(data, schema=IWords), {"words": ["ä", "ö"]})

    def test_skip_decoding_primary_fields(self):
        data = {"words": ["ä", "ö"]}
        self.assertEqual(decode(data, schema=IPrimary), {})



from plone.tiles.data import encode
from plone.tiles.interfaces import ITileDataManager
from plone.tiles.interfaces import ITileType
from urllib import parse
from zope.annotation import IAnnotations
from zope.component import getMultiAdapter
from zope.component import queryUtility
from zope.traversing.browser.absoluteurl import AbsoluteURL
from zope.traversing.browser.interfaces import IAbsoluteURL


_safe = "@+"


class BaseTileAbsoluteURL(AbsoluteURL):
    """Convenience base class"""

    def __str__(self):
        tile = self.context
        request = self.request

        tid = tile.id
        name = tile.__name__
        context = tile.__parent__

        if name is None or context is None:
            raise TypeError("Insufficient context to determine URL")

        tileFragment = "@@" + parse.quote(name.encode("utf-8"), _safe)
        if tid:
            tileFragment += "/" + parse.quote(tid.encode("utf-8"), _safe)

        absolute_url = getMultiAdapter((context, request), IAbsoluteURL)
        try:
            tileFragment = "/".join([str(absolute_url), tileFragment])
        except TypeError:  # Not enough context to get URL information
            pass

        return tileFragment

    def breadcrumbs(self):
        tile = self.context
        request = self.request

        tid = tile.id
        name = tile.__name__
        context = tile.__parent__

        tileFragment = "@@" + parse.quote(name.encode("utf-8"), _safe)
        if tid:
            tileFragment += "/" + parse.quote(tid.encode("utf-8"), _safe)

        base = tuple(getMultiAdapter((context, request), IAbsoluteURL).breadcrumbs())
        base += (
            {
                "name": name,
                "url": "/".join([base[-1]["url"], tileFragment]),
            },
        )

        return base


class TransientTileAbsoluteURL(BaseTileAbsoluteURL):
    """Absolute URL for a transient tile. Includes the tile traverser and
    tile data encoded in the query string.
    """

    def __str__(self):
        url = super().__str__()
        manager = ITileDataManager(self.context)

        # Transient looking tile with id is only really transient
        # if it caches its decoded query data in request annotations
        transient = manager.storage == IAnnotations(self.request)

        # When transient looking tile with id is not really transient,
        # its data should not be encoded into query string
        if self.context.id and not transient:
            return url

        # All tiles don't need / have configuration data at all.
        data = manager.get()
        if not data:
            return url

        # But when configuration data is really read from query string
        # and not persisted, it should also be kept in query string
        tileType = queryUtility(ITileType, name=self.context.__name__)
        if tileType is not None and tileType.schema is not None:
            if "?" in url:
                url += "&" + encode(data, tileType.schema)
            else:
                url += "?" + encode(data, tileType.schema)
        return url


class PersistentTileAbsoluteURL(BaseTileAbsoluteURL):
    """Absolute URL for a persistent tile. Includes the tile traverser, but no
    tile data encoded in the query string.
    """



from plone.tiles.interfaces import IFieldTypeConverter
from zope.interface import implementer


@implementer(IFieldTypeConverter)
class NoConverter:

    def __init__(self, field):
        self.field = field

    token = None


class TextConverter(NoConverter):
    token = "text"


class LongConverter(NoConverter):
    token = "long"


class FloatConverter(NoConverter):
    token = "float"


class BoolConverter(NoConverter):
    token = "boolean"


class TupleConverter(NoConverter):
    token = "tuple"


class ListConverter(NoConverter):
    token = "list"


class DictConverter(NoConverter):
    token = "record"



from plone.testing import layered
from plone.tiles.testing import PLONE_TILES_INTEGRATION_TESTING

import doctest
import re
import unittest


class Py23DocChecker(doctest.OutputChecker):
    def check_output(self, want, got, optionflags):
        want = re.sub("u'(.*?)'", "'\\1'", want)
        return doctest.OutputChecker.check_output(self, want, got, optionflags)


def test_suite():
    return unittest.TestSuite(
        (
            layered(
                doctest.DocFileSuite(
                    "../tiles.rst",
                    "../directives.rst",
                    "../data.rst",
                    "../esi.rst",
                    optionflags=doctest.ELLIPSIS,
                    checker=Py23DocChecker(),
                ),
                layer=PLONE_TILES_INTEGRATION_TESTING,
            ),
        )
    )



from plone.tiles.interfaces import IPersistentTile
from plone.tiles.interfaces import ITile
from plone.tiles.interfaces import ITileDataManager
from Products.Five import BrowserView
from zExceptions import Forbidden
from zope.component import queryMultiAdapter
from zope.interface import implementer
from zope.traversing.browser.absoluteurl import absoluteURL


@implementer(ITile)
class Tile(BrowserView):
    """Basic implementation of a transient tile. Subclasses should override
    __call__ or set an 'index' variable to point to a view page template file.

    The tile is basically a browser view, with the following enhancements:

    * The attribute `data` can be used to read the tile data, as returned by
      `ITileDataManager(tile).get()`. This value is cached when it is first
      read.
    * The attribute `url` can be used to obtain the tile's URL, including the
      id specifier and any data associated with a transient tile. Again, the
      return value is cached after the first access.
    * The class implements __getitem__() to set the tile id from the traversal
      sub-path, as well as to allow views to be looked up. This is what allows
      a URL like `http://.../@@example.tile/foo` to result in a tile with id
      `foo`.
    """

    __cachedData = None
    __cachedURL = None

    id = None

    def __getitem__(self, name):

        # If we haven't set the id yet, do that first
        if self.id is None:
            self.id = name

            # This is pretty stupid, but it's required to keep the ZPublisher
            # happy in Zope 2. It doesn't normally check for docstrings on
            # views, but it does check for them on sub-objects obtained via
            # __getitem__.

            if self.__doc__ is None:
                self.__doc__ = "For Zope 2, to keep the ZPublisher happy"

            # Note: X-Tile-Url was added to make it easier for editor to know
            # the URL of a new tile after receiving the redirected response
            # from a tile form. That's why it's only set for customizable tiles
            # (tiles with id).
            if self.id is not None:
                self.request.response.setHeader("X-Tile-Url", self.url)

            return self

        # Also allow views on tiles even without @@.
        viewName = name
        if viewName.startswith("@@"):
            viewName = name[2:]
        view = queryMultiAdapter((self, self.request), name=viewName)
        if view is not None:
            view.__parent__ = self
            view.__name__ = viewName
            return view

        raise KeyError(name)

    def browserDefault(self, request):
        """By default, tiles render themselves with no browser-default view"""
        return self, ()

    def publishTraverse(self, request, name):
        """Ensure that publish-traversal uses the same semantics as
        __getitem__.
        """
        return self[name]

    def __call__(self, *args, **kwargs):
        if getattr(self, "index", None) is None:
            raise NotImplementedError(
                'Override __call__ or set a class variable "index" to point '
                "to a view page template file"
            )
        return self.index(*args, **kwargs)

    @property
    def data(self):
        if self.__cachedData is None:
            reader = ITileDataManager(self)
            self.__cachedData = reader.get()
        return self.__cachedData

    @property
    def url(self):
        return absoluteURL(self, self.request)


@implementer(IPersistentTile)
class PersistentTile(Tile):
    """Base class for persistent tiles. Identical to `Tile`, except that the
    data dict is never serialized with the URL.
    """

    def absolute_url(self):
        # the "absolute_url" adapter for persistent tiles never has query parameters.
        # We can safely return self.url here.
        return self.url


class TileThemingTransform:
    """Disable plone.app.theming for tile responses"""

    order = 8800

    def __init__(self, published, request):
        self.published = published
        self.request = request

    def transform(self, result, encoding):
        self.request.response.setHeader("X-Theme-Disabled", "1")
        return None

    def transformBytes(self, result, encoding):
        return self.transform(result, encoding)

    def transformUnicode(self, result, encoding):
        return self.transform(result, encoding)

    def transformIterable(self, result, encoding):
        return self.transform(result, encoding)


class TileProtectTransform:
    """Replacement transform for plone.protect's ProtectTransform, to drop
    X-Tile-Url-header from unauthorized responses and disable the default
    ProtectTransform for authorized responses (to avoid causing issues
    like extra protect.js-injections for tile editors)
    """

    order = 9000

    def __init__(self, published, request):
        self.published = published
        self.request = request
        try:
            from plone.protect.auto import ProtectTransform

            self.protect = ProtectTransform(published, request)
        except ImportError:
            self.protect = None

    def transform(self, result, encoding):
        from plone.protect import CheckAuthenticator

        CheckAuthenticator(self.request)
        return None

    def transformBytes(self, result, encoding):
        try:
            return self.transform(result, encoding)
        except Forbidden:
            if "x-tile-url" in self.request.response.headers:
                del self.request.response.headers["x-tile-url"]
            if self.protect is not None:
                return self.protect.transformBytes(result, encoding)
            else:
                return None

    def transformUnicode(self, result, encoding):
        try:
            return self.transform(result, encoding)
        except Forbidden:
            if "x-tile-url" in self.request.response.headers:
                del self.request.response.headers["x-tile-url"]
            if self.protect is not None:
                return self.protect.transformUnicode(result, encoding)
            else:
                return None

    def transformIterable(self, result, encoding):
        try:
            return self.transform(result, encoding)
        except Forbidden:
            if "x-tile-url" in self.request.response.headers:
                del self.request.response.headers["x-tile-url"]
            if self.protect is not None:
                return self.protect.transformIterable(result, encoding)
            else:
                return None



from plone.tiles.interfaces import ITileType
from plone.tiles.tile import Tile
from plone.tiles.type import TileType
from Products.Five.browser.metaconfigure import page
from zope import schema
from zope.component.zcml import utility
from zope.configuration.exceptions import ConfigurationError
from zope.configuration.fields import GlobalInterface
from zope.configuration.fields import GlobalObject
from zope.configuration.fields import MessageID
from zope.configuration.fields import Path
from zope.interface import Interface
from zope.publisher.interfaces.browser import IDefaultBrowserLayer
from zope.security.zcml import Permission


class ITileDirective(Interface):
    """Directive which registers a new type of tile"""

    name = schema.DottedName(
        title="Name",
        description="A unique, dotted name for the tile",
        required=True,
    )

    title = MessageID(
        title="Title",
        description="A user friendly title, used when configuring the tile",
        required=False,
    )

    description = MessageID(
        title="Description",
        description="A longer summary of the tile's purpose and function",
        required=False,
    )

    icon = MessageID(
        title="Icon",
        description="Image that represents tile purpose and function",
        required=False,
    )

    add_permission = Permission(
        title="Add permission",
        description="Name of the permission required to instantiate " "this tile",
        required=False,
    )

    edit_permission = Permission(
        title="Edit permission",
        description="Name of the permission required to edit this tile",
        required=False,
    )

    delete_permission = Permission(
        title="Delete permission",
        description="Name of the permission required to delete this tile",
        required=False,
    )

    schema = GlobalInterface(
        title="Configuration schema for the tile",
        description="This is used to create standard add/edit forms",
        required=False,
    )

    for_ = GlobalObject(
        title="The interface or class this tile is available for",
        required=False,
    )

    layer = GlobalInterface(title="The layer the tile is available for", required=False)

    class_ = GlobalObject(
        title="Class", description="Class implementing this tile", required=False
    )

    template = Path(
        title="The name of a template that renders this tile",
        description="Refers to a file containing a page template",
        required=False,
    )

    permission = Permission(
        title="View permission",
        description="Name of the permission required to view this item",
        required=True,
    )


def tile(
    _context,
    name,
    title=None,
    description=None,
    icon=None,
    add_permission=None,
    edit_permission=None,
    delete_permission=None,
    schema=None,
    for_=None,
    layer=None,
    class_=None,
    template=None,
    permission=None,
):
    """Implements the <plone:tile /> directive"""
    if (
        title is not None
        or description is not None
        or icon is not None
        or add_permission is not None
        or schema is not None
    ):
        if title is None or add_permission is None:
            raise ConfigurationError(
                "When configuring a new type of tile, 'title' and "
                "'add_permission' are required"
            )
        type_ = TileType(
            name,
            title,
            add_permission,
            permission,
            edit_permission=edit_permission,
            delete_permission=delete_permission,
            description=description,
            icon=icon,
            schema=schema,
        )

        utility(_context, provides=ITileType, component=type_, name=name)

    if (
        for_ is not None
        or layer is not None
        or class_ is not None
        or template is not None
    ):
        if class_ is None and template is None:
            raise ConfigurationError(
                "'class' or 'template' must be given when configuring a tile."
            )

        if for_ is None:
            for_ = Interface
        if layer is None:
            layer = IDefaultBrowserLayer

        if class_ is None:
            class_ = Tile

        page(
            _context,
            name=name,
            permission=permission,
            for_=for_,
            layer=layer,
            template=template,
            class_=class_,
        )



from plone.testing import Layer
from plone.testing import z2
from plone.testing import zca
from plone.tiles import PersistentTile
from plone.tiles import Tile
from zope import schema
from zope.configuration import xmlconfig
from zope.interface import Interface


# For directive tests


class IDummySchema(Interface):
    foo = schema.TextLine(title="Foo")


class IDummyContext(Interface):
    pass


class IDummyLayer(Interface):
    pass


class DummyTile(Tile):

    def __call__(self):
        return "dummy"


class DummyTileWithTemplate(PersistentTile):
    pass


class PloneTiles(Layer):
    defaultBases = (z2.STARTUP,)

    def setUp(self):
        self["configurationContext"] = context = zca.stackConfigurationContext(
            self.get("configurationContext")
        )
        import zope.annotation

        xmlconfig.file("configure.zcml", zope.annotation, context=context)
        import plone.tiles

        xmlconfig.file("configure.zcml", plone.tiles, context=context)

    def tearDown(self):
        del self["configurationContext"]


PLONE_TILES_FIXTURE = PloneTiles()


PLONE_TILES_INTEGRATION_TESTING = z2.IntegrationTesting(
    bases=(PLONE_TILES_FIXTURE,), name="PloneTiles:Functional"
)



from plone.tiles.interfaces import ITileType
from zope.interface import implementer


@implementer(ITileType)
class TileType:
    """A utility that describes a type of tile"""

    def __init__(
        self,
        name,
        title,
        add_permission,
        view_permission,
        edit_permission=None,
        delete_permission=None,
        description=None,
        icon=None,
        schema=None,
    ):

        if delete_permission is None:
            delete_permission = add_permission

        if edit_permission is None:
            edit_permission = add_permission

        self.__name__ = name
        self.title = title
        self.add_permission = add_permission
        self.edit_permission = edit_permission
        self.view_permission = view_permission
        self.delete_permission = delete_permission
        self.description = description
        self.icon = icon
        self.schema = schema

    def __repr__(self):
        return f"<TileType {self.__name__} ({self.title})>"



from zope.interface import Interface
from zope.interface.common.mapping import IMapping
from zope.interface.interfaces import IInterface
from zope.publisher.interfaces.browser import IBrowserView

import zope.schema


ESI_HEADER = "X-ESI-Enabled"
ESI_HEADER_KEY = "HTTP_" + ESI_HEADER.replace("-", "_").upper()


class ITileType(Interface):
    """A utility that describes a type of tile"""

    __name__ = zope.schema.DottedName(title="Tile name (same as utility name)")

    title = zope.schema.TextLine(title="Title")

    description = zope.schema.Text(title="Description", required=False)

    icon = zope.schema.Text(title="Icon", required=False)

    add_permission = zope.schema.Id(title="Zope 3 IPermission utility name")

    schema = zope.schema.Object(
        title="Tile schema",
        description="Describes configurable data for this tile and allows a "
        "form to be rendered to edit it. Set to None if the tile "
        "has no configurable schema",
        schema=IInterface,
        required=False,
    )


class IBasicTile(IBrowserView):
    """A tile is a publishable resource that can be inserted into a site or
    page layout.

    The tile should be a named multi adapter on (<context>, <layer>)
    providing IBasicTile.

    It will normally be traversed to like this::

        http://localhost:8080/plone-site/object/@@my.tile/tile1

    In this case:

    * The tile context is the content object at /plone-site/object.
    * The ``__name__`` of the tile instance is 'my.tile'
    * The ``id`` of the tile instance is 'tile1'
    * The ``url`` of the tile instance is the URL as above
    """

    __name__ = zope.schema.DottedName(
        title="The name of the type of this tile",
        description="This should be a dotted name prefixed with the "
        "package that defined the tile",
    )

    id = zope.schema.DottedName(
        title="Tile instance id",
        description="The id is normally set using sub-path traversal"
        "A given tile type may be used multiple times on "
        "the same page, each with a unique id. The id must "
        "be unique even across multiple layouts for the "
        "same context.",
    )


class ITile(IBasicTile):
    """A tile with some data (probably from a query string)."""

    data = zope.schema.Dict(
        title="The tile's configuration data",
        description="This attribute cannot be set, but the dictionary may "
        "be updated",
        key_type=zope.schema.Id(title="The data element name"),
        value_type=zope.schema.Field(title="The value"),
        required=True,
        readonly=True,
        default={},
    )

    url = zope.schema.URI(
        title="Tile URL",
        description="This is the canonical URL for the tile. In the "
        "case of transient tiles with data, this may "
        "include a query string with parameters. Provided "
        "that the `id` attribute is set, it will also "
        "include a sub-path with this in it.",
    )


class IPersistentTile(ITile):
    """A tile with full-blown persistent data (stored in annotations)."""


class IESIRendered(Interface):
    """Marker interface for tiles which are to be rendered via ESI.

    Two corresponding views, @@esi-body and @@esi-head, will be made available
    on the tile itself. This will return the children of the <head /> or
    <body /> of the tile, respectively.

    Thus, a tile marked with this interface may be replaced with an ESI
    instruction like::

        <esi:include src="./@@my.tile/tile-id/@@esi-body" />

    When fetched, this placeholder will be replaced by the body of the tile.
    """


class ITileDataManager(Interface):
    """Support for getting and setting tile data dicts.

    This is an adapter on a tile. The tile's id must be set.
    """

    def get():
        """Get a dictionary with tile data for this tile. The dictionary is
        disconnected from the underlying storage.
        """

    def set(data):
        """Persist the given data dict."""

    def delete():
        """Delete the data record for this tile."""


class ITileDataContext(Interface):
    """Indirection to help determine where persistent tiles store their data.

    This is a multi-adapter on ``(context, request, tile)``. The context and
    request are the same as ``tile.context`` and ``tile.request``, but these
    discriminators allow the data context to be customised depending on
    the context or request.

    The default implementation simply returns ``tile.context``. That must
    be annotatable for the default tile data storage adapter and
    persistent tile ``ITileDataManager`` to work.
    """


class ITileDataStorage(IMapping):
    """Indirection to help determine how persistent tiles store their data.

    This is a multi-adapter on ``(context, request, tile)``. The context and
    request are the same as ``tile.context`` and ``tile.request``, but these
    discriminators allow the data context to be customised depending on
    the context or request.

    The default implementation simply returns the configured zope.annotation
    storage for the given context.

    The adapter is expected to provide IMapping interface and be accessed
    by tile data managers similarly to zope.annotation storage.
    """


class IFieldTypeConverter(Interface):
    """Field type converter for querystring parameters for Zope."""

    token = zope.schema.TextLine(
        title="Token",
        description="""
                                 String parameter appended to the field id
                                 for the Zope Publisher to cast it.
                                 """,
    )



from plone.supermodel.directives import MetadataListDirective


IGNORE_QUERYSTRING_KEY = "plone.tiles.ignore_querystring"


class ignore_querystring(MetadataListDirective):
    """Directive used to create fieldsets"""

    key = IGNORE_QUERYSTRING_KEY

    def factory(self, name):
        return [name]



from setuptools import find_packages
from setuptools import setup

import os


def read(*path):
    filename = os.path.join(*path)
    with open(filename) as myfile:
        return myfile.read() + "\n"


version = "3.0.2"

setup(
    name="plone.tiles",
    version=version,
    description="APIs for managing tiles",
    long_description=read("README.rst")
    + read("plone", "tiles", "tiles.rst")
    + read("plone", "tiles", "directives.rst")
    + read("plone", "tiles", "esi.rst")
    + read("CHANGES.rst"),
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Web Environment",
        "Framework :: Plone",
        "Framework :: Plone :: 6.0",
        "Framework :: Plone :: 6.1",
        "Framework :: Plone :: 6.2",
        "Framework :: Plone :: Core",
        "License :: OSI Approved :: GNU General Public License v2 (GPLv2)",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    keywords="plone tiles",
    author="Martin Aspeli",
    author_email="optilude@gmail.com",
    url="https://github.com/plone/plone.tiles",
    license="GPL version 2",
    packages=find_packages(),
    namespace_packages=["plone"],
    include_package_data=True,
    zip_safe=False,
    python_requires=">=3.8",
    install_requires=[
        "plone.subrequest",
        "setuptools",
        "zope.annotation",
        "zope.component",
        "zope.configuration",
        "zope.interface",
        "zope.publisher",
        "zope.schema",
        "zope.security",
        "zope.traversing",
        "plone.protect",
        "plone.supermodel",
        "Zope",
    ],
    extras_require={
        "test": [
            "plone.testing [zca, z2]",
            "plone.rfc822",
        ],
    },
)



from importlib.metadata import distribution
from importlib.metadata import PackageNotFoundError
from persistent.dict import PersistentDict
from plone.subrequest import ISubRequest
from plone.tiles.directives import IGNORE_QUERYSTRING_KEY
from plone.tiles.interfaces import IFieldTypeConverter
from plone.tiles.interfaces import IPersistentTile
from plone.tiles.interfaces import ITile
from plone.tiles.interfaces import ITileDataContext
from plone.tiles.interfaces import ITileDataManager
from plone.tiles.interfaces import ITileDataStorage
from plone.tiles.interfaces import ITileType
from urllib import parse
from zope.annotation.interfaces import IAnnotations
from zope.component import adapter
from zope.component import getMultiAdapter
from zope.component import queryUtility
from zope.interface import implementer
from zope.interface import Interface
from zope.interface.interfaces import ComponentLookupError
from zope.schema import getFields
from zope.schema import getFieldsInOrder
from zope.schema.interfaces import ISequence

import json
import logging


try:
    distribution("plone.rfc822")
except PackageNotFoundError:
    HAS_RFC822 = False
else:
    from plone.rfc822.interfaces import IPrimaryField

    HAS_RFC822 = True


ANNOTATIONS_KEY_PREFIX = "plone.tiles.data"
LOGGER = logging.getLogger("plone.tiles")


@adapter(ITile)
@implementer(ITileDataManager)
def transientTileDataManagerFactory(tile):
    if tile.request.get("X-Tile-Persistent"):
        return PersistentTileDataManager(tile)
    else:
        return TransientTileDataManager(tile)


class BaseTileDataManager:

    def get_default_request_data(self):
        """
        from request form
        """
        # try to use a '_tiledata' parameter in the request
        if "_tiledata" in self.tile.request.form:
            data = json.loads(self.tile.request.form["_tiledata"])
        elif self.tileType is None or self.tileType.schema is None:
            data = self.tile.request.form.copy()
        else:
            # Try to decode the form data properly if we can
            try:
                data = decode(
                    self.tile.request.form,
                    self.tileType.schema,
                    missing=True,
                    primary=True,
                )
            except (
                ValueError,
                UnicodeDecodeError,
            ):
                LOGGER.exception("Could not convert form data to schema")
                return self.data.copy()
        # we're assuming this data is potentially unsafe so we need to check
        # the ignore querystring field setting

        # before we start, we allow it for sub-requests since in this case,
        # the input is safe and we can trust it
        if ISubRequest.providedBy(self.tile.request):
            return data

        # first off, we only care to filter if it is a GET request
        if getattr(self.tile.request, "REQUEST_METHOD", "GET") != "GET":
            return data

        # now, pay attention to schema hints for form data
        if self.tileType is not None and self.tileType.schema is not None:
            for name in (
                self.tileType.schema.queryTaggedValue(IGNORE_QUERYSTRING_KEY) or []
            ):
                if name in data:
                    del data[name]

        return data


@adapter(ITile)
@implementer(ITileDataManager)
class TransientTileDataManager(BaseTileDataManager):
    """A data manager for transient tile data, which reads data from the
    request query string.
    """

    def __init__(self, tile):
        self.tile = tile
        self.tileType = queryUtility(ITileType, name=tile.__name__)

        self.context = getMultiAdapter(
            (tile.context, tile.request, tile), ITileDataContext
        )
        self.storage = getMultiAdapter(
            (self.context, tile.request, tile), ITileDataStorage
        )

        if IAnnotations.providedBy(self.storage):
            self.key = ".".join([ANNOTATIONS_KEY_PREFIX, str(tile.id)])
        else:
            self.key = str(tile.id)

    @property
    def annotations(self):  # BBB for < 0.7.0 support
        return self.storage

    def get(self):
        # use explicitly set data (saved as annotation on the request)
        if self.key in self.storage:
            data = dict(self.storage[self.key])

            if self.tileType is not None and self.tileType.schema is not None:
                for name, field in getFields(self.tileType.schema).items():
                    if name not in data:
                        data[name] = field.missing_value
        # fall back to the copy of request.form object itself
        else:
            data = self.get_default_request_data()

        return data

    def set(self, data):
        self.storage[self.key] = data

    def delete(self):
        if self.key in self.storage:
            self.storage[self.key] = {}


@adapter(IPersistentTile)
@implementer(ITileDataManager)
class PersistentTileDataManager(BaseTileDataManager):
    """A data reader for persistent tiles operating on annotatable contexts.
    The data is retrieved from an annotation.
    """

    def __init__(self, tile):
        self.tile = tile
        self.tileType = queryUtility(ITileType, name=tile.__name__)

        self.context = getMultiAdapter(
            (tile.context, tile.request, tile), ITileDataContext
        )
        self.storage = getMultiAdapter(
            (self.context, tile.request, tile), ITileDataStorage
        )

        if IAnnotations.providedBy(self.storage):
            self.key = ".".join([ANNOTATIONS_KEY_PREFIX, str(tile.id)])
        else:
            self.key = str(tile.id)

    @property
    def annotations(self):  # BBB for < 0.7.0 support
        return self.storage

    def get(self):
        data = self.get_default_request_data()
        data.update(dict(self.storage.get(self.key, {})))
        if self.tileType is not None and self.tileType.schema is not None:
            for name, field in getFields(self.tileType.schema).items():
                if name not in data:
                    data[name] = field.missing_value
        return data

    def set(self, data):
        self.storage[self.key] = PersistentDict(data)

    def delete(self):
        if self.key in self.storage:
            del self.storage[self.key]


@implementer(ITileDataContext)
@adapter(Interface, Interface, ITile)
def defaultTileDataContext(context, request, tile):
    return tile.context


@implementer(ITileDataStorage)
@adapter(Interface, Interface, ITile)
def defaultTileDataStorage(context, request, tile):
    if tile.request.get("X-Tile-Persistent"):
        return defaultPersistentTileDataStorage(context, request, tile)
    else:
        return IAnnotations(tile.request, tile.request.form)


@implementer(ITileDataStorage)
@adapter(Interface, Interface, IPersistentTile)
def defaultPersistentTileDataStorage(context, request, tile):
    return IAnnotations(context)


# Encoding


def map_to_pairs(encoded_name, value):
    """Given an encoded basename, e.g. "foo:record" or "foo:record:list" and
    a dictionary value, yields (encoded_name, value) pairs to be included
    in the final encode.
    """
    prefix, postfix = encoded_name.split(":", 1)
    postfix = postfix.replace("record:list", "records")

    def guess_type(v):
        if isinstance(v, str):
            return ""
        if isinstance(v, bool):
            return ":boolean"
        if isinstance(v, int):
            return ":int"
        if isinstance(v, float):
            return ":float"
        return ""

    for item_name, item_value in value.items():
        if isinstance(item_value, str):
            item_value = item_value.encode("utf-8")

        if isinstance(item_value, list) or isinstance(item_value, tuple):
            for item_subvalue in item_value:
                marshall_type = guess_type(item_subvalue)
                if isinstance(item_subvalue, bool):
                    item_subvalue = item_subvalue and "1" or ""
                elif isinstance(item_subvalue, str):
                    item_subvalue = item_subvalue.encode("utf-8")
                encoded_name = "{}.{}{}:list:{}".format(
                    prefix, item_name, marshall_type, postfix
                )
                yield encoded_name, item_subvalue
        else:
            marshall_type = guess_type(item_value)
            if isinstance(item_value, bool):
                item_value = item_value and "1" or ""
            elif isinstance(item_value, str):
                item_value = item_value.encode("utf-8")
            encoded_name = "{:s}.{:s}{:s}:{:s}".format(
                prefix, item_name, marshall_type, postfix
            )
            yield encoded_name, item_value


def encode(data, schema, ignore=()):
    """Given a data dictionary with key/value pairs and schema, return an
    encoded query string. This is similar to urllib.urlencode(), but field
    names will include the appropriate field type converters, e.g. an int
    field will be encoded as fieldname:int=123. Fields not found in the data
    dict will be ignored, and items in the dict not in the schema will also
    be ignored. Additional fields to ignore can be passed with the 'ignore'
    parameter. If any fields cannot be converted, a ComponentLookupError
    will be raised.
    """

    encode = []

    for name, field in getFieldsInOrder(schema):
        if HAS_RFC822 and IPrimaryField.providedBy(field):
            continue

        if name in ignore or name not in data:
            continue

        converter = IFieldTypeConverter(field, None)
        if converter is None:
            raise ComponentLookupError(
                f"Cannot URL encode {name} of type {field.__class__}"
            )

        encoded_name = name
        if converter.token:
            encoded_name = ":".join([name, converter.token])

        value = data[name]
        if value is None:
            continue
        elif isinstance(value, str):
            value = value.encode("utf-8")

        if ISequence.providedBy(field):
            value_type_converter = IFieldTypeConverter(field.value_type, None)
            if value_type_converter is None:
                raise ComponentLookupError(
                    "Cannot URL encode value type for {} of type "
                    "{} : {}".format(name, field.__class__, field.value_type.__class__)
                )

            if value_type_converter.token:
                encoded_name = ":".join(
                    [name, value_type_converter.token, converter.token]
                )

            for item in value:

                if isinstance(item, bool):
                    item = item and "1" or ""
                elif isinstance(item, str):
                    item = item.encode("utf-8")

                if isinstance(item, dict):
                    encode.extend(map_to_pairs(encoded_name, item))
                else:
                    encode.append(
                        (
                            encoded_name,
                            item,
                        )
                    )

        else:
            # The :bool converter just does bool() value, but urlencode() does
            # str() on the object. The result is False => 'False' => True :(
            if isinstance(value, bool):
                value = value and "1" or ""

            if isinstance(value, dict):
                encode.extend(map_to_pairs(encoded_name, value))
            else:
                encode.append((encoded_name, value))

    return parse.urlencode(encode)


# Decoding


def decode(data, schema, missing=True, primary=False):
    """Decode a data dict according to a schema. The returned dictionary will
    contain only keys matching schema names, and will force type values
    appropriately.

    This function is only necessary because the encoders used by encode()
    are not sufficiently detailed to always return the exact type expected
    by a field, e.g. resulting in ascii/unicode discrepancies.

    If missing is True, fields that are in the schema but not in the data will
    be set to field.missing_value. Otherwise, they are ignored.

    If primary is True, also fields that are marged as primary fields are
    decoded from the data. (Primary fields are not decoded by default,
    because primary field are mainly used for rich text or binary fields
    and data is usually parsed from query string with length limitations.)
    """

    decoded = {}

    for name, field in getFields(schema).items():
        if not primary and HAS_RFC822 and IPrimaryField.providedBy(field):
            continue

        if name not in data:
            if missing:
                decoded[name] = field.missing_value
            continue

        value = data[name]
        if value is None:
            continue

        field_type = field._type
        if isinstance(
            field_type,
            (
                tuple,
                list,
            ),
        ):
            field_type = field_type[-1]

        if ISequence.providedBy(field):
            converted = []

            value_type_field_type = field.value_type._type
            if isinstance(
                value_type_field_type,
                (
                    tuple,
                    list,
                ),
            ):
                value_type_field_type = value_type_field_type[-1]

            for item in value:
                if field.value_type._type and not isinstance(
                    item, field.value_type._type
                ):
                    item = value_type_field_type(item)
                converted.append(item)

            value = converted
        elif isinstance(value, (tuple, list)) and value:
            value = value[0]

        if isinstance(value, bytes):
            value = value.decode("utf-8")

        if field._type is not None and not isinstance(value, field._type):
            value = field_type(value)

        decoded[name] = value

    return decoded
